#include "../assets.hpp"
#include "font_writer.hpp"
#include <PSX/GPU/gpu.hpp>
#include <PSX/GPU/gpu_primitives.hpp>

namespace FontWriter {
    using namespace JabyEngine;
    static GPU::TexPage::Linked CharTexPage[2] = {
        GPU::TexPage({Assets::FontTIM.get_texture_x(), Assets::FontTIM.get_texture_y()}, GPU::TexturePageColor::$4bit).linked(),
        GPU::TexPage({Assets::FontTIM.get_texture_x(), Assets::FontTIM.get_texture_y()}, GPU::TexturePageColor::$4bit).linked()
    };
    static GPU::SPRT_16::Linked  TextBuffer[2][TextBufferSize];
    static GPU::Link*            TextLink;
    static GPU::SPRT_16::Linked* TextPtr;

    static void reset_links() {
        TextLink = &CharTexPage[GPU::Display::current_id];
        TextPtr  = TextBuffer[GPU::Display::current_id];
    }

    void setup() {
        reset_links();
        for(auto& char_array : TextBuffer) {
            for(auto& single_char : char_array) {
                single_char->set_identitiy();
                single_char->clut = GPU::PageClut(Assets::FontTIM.get_clut_x(), Assets::FontTIM.get_clut_y());
            }
        }
    }

    Position write(Position pos, const char* text) {
        static const auto parse_esc = [](const char* text, const GPU::Color24& color) -> pair<const char*, GPU::Color24> {
            static const auto char_to_color = [](char number) constexpr -> uint8_t {
                return (1 << (number - '0')) - 1;
            };
            if(text[0] != '[' || text[2] != ';' || text[4] != ';' || text[6] != 'm') {
                return {text, color};
            }

            return {text + 7, GPU::Color24(char_to_color(text[1]), char_to_color(text[3]), char_to_color(text[5]))};
        };
        const auto* cur_text_end = &TextBuffer[GPU::Display::current_id][TextBufferSize];
        const auto  org_x        = pos.x;
        auto        font_color   = GPU::Color24::Grey();

        while(TextPtr < cur_text_end) {
            const char cur_char = *text;
            text++;

            (*TextPtr)->position = pos;
            switch(cur_char) {
                case '\0':
                    goto end;
    
                case '\n':
                    pos.y += 16;
                    pos.x  = org_x;
                    continue;
                
                case '\x1b':
                    tie(text, font_color) = parse_esc(text, font_color);
                    continue;

                case 'i':
                case '!':
                    pos.x += 12;
                    break;

                default:
                    pos.x += 16;
            }

            const uint8_t char_id = cur_char - '!';
            (*TextPtr)->page     = {static_cast<uint8_t>((char_id & 0xF) << 4), static_cast<uint8_t>((char_id >> 4) << 4)};
            (*TextPtr)->color    = font_color;

            TextLink = &TextLink->concat(*TextPtr);
            TextPtr++;
        }

    end:
        return pos;
    }

    void render() {
        const auto render_id = GPU::Display::current_id ^ 1;

        TextLink->terminate();
        GPU::render(CharTexPage[render_id]);
        reset_links();
    }
}
